// InputPort.cpp: implementation of the InputPort class.
//
//////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "InputPort.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

InputPort::InputPort()
{
  m_prevTime = 0;
}

InputPort::~InputPort()
{
  VectorPEventIt pEvtIt;

  for ( pEvtIt=m_vecPEvent.begin(); pEvtIt!=m_vecPEvent.end(); pEvtIt++)
  {
    delete *pEvtIt;
  }
  /*
  size_t i;
  for (i=0; i<m_vecIntTotal.size(); i++)
  {
    cout << i << '\t' << m_vecIntTotal[i] << endl;
  }
  */
}

void InputPort::InitVc(ifstream &cfgFile)
{
  char line[255];
  DWORD id, qid;
  double peakRate, avgRate, tknBckt, tknRate;
  int repeat;
  int lineNo;
  int i;
  double beginTimeSec;
  double lengthSec;
  PVC pvc, pNewVc;
  double avgOn;
  DWORD seedOn, seedOff;
  int type;
  istrstream *pis;

  id = 0;
  lineNo = 0;
  while (!cfgFile.eof()) // read until eof
  {
    ReadLine(cfgFile,line,254);
    if (IsBlank(line) && cfgFile.eof())
      break;

    pis = new istrstream(line);
    *pis >> qid >> peakRate >> avgRate 
         >> tknBckt >> tknRate >> type 
         >> avgOn >> seedOn >> seedOff
         >> beginTimeSec >> lengthSec >> repeat;
    delete pis;
    if (repeat<1)
      Error("Config file %d, repeat error!", line);

    if ((type<0)||(type>2))
      Error("Error type");

    if (tknRate==0)
      Error("Error token rate");

    pvc = new VC(-1,qid,beginTimeSec, lengthSec, 
                 peakRate, avgRate, tknRate, tknBckt, type,
                 avgOn, seedOn, seedOff);
    for (i=0; i<repeat; i++)
    {
      pvc->m_id = id++;
      m_vecIntTotal.push_back(0);
      pvc->SetSeed(seedOn+i*i, seedOff+i*i);
      // init the event heap
      pNewVc = m_vcHash.AddVc(pvc);
      AddEvent(pNewVc->GetNextEvent());
    }
    delete pvc;
  }
  if (id==0)
    Error("No flows defined in config file!");
}

BOOL InputPort::LaterEvent(PEvent &x, PEvent &y)
{
  if (x->pPacket->m_time > y->pPacket->m_time)
    return TRUE;
  else if (x->pPacket->m_time == y->pPacket->m_time)
  {
    //if (x->pPacket->m_seq > y->pPacket->m_seq)
    if (x->pPacket->m_vcId > y->pPacket->m_vcId)
      return TRUE; 
    else
      return FALSE;
  }
  else
    return FALSE;
}

PEvent InputPort::GetNextEvent()
{
  PEvent pEvent;
  PEvent pNewEvent;
  ListVCIt pVc;

  if (m_vecPEvent.size()==0) // all flows finish
    return NULL;
  pop_heap(m_vecPEvent.begin(), m_vecPEvent.end(), LaterEvent);
  pEvent = m_vecPEvent.back(); // get the next event with the smallest time
  m_vecPEvent.pop_back();

  if (pEvent->what != EVT_VC_STOP)
  { // put the new event in the heap
    pVc = m_vcHash.GetVcFromHash(pEvent->pPacket->m_vcId);
    pNewEvent = pVc->GetNextEvent();
    assert(pNewEvent->pPacket->m_time >= pEvent->pPacket->m_time);
    m_vecPEvent.push_back(pNewEvent);
    push_heap(m_vecPEvent.begin(), m_vecPEvent.end(), LaterEvent);
    m_vecIntTotal[pVc->m_id]++;
  }
  //ofLog1 << *(pEvent->pPacket);
  assert(m_prevTime<=pEvent->pPacket->m_time);
  m_prevTime = pEvent->pPacket->m_time;
  return pEvent;
}

void InputPort::AddEvent(PEvent pEvent)
{
  m_vecPEvent.push_back(pEvent);
  push_heap(m_vecPEvent.begin(), m_vecPEvent.end(), LaterEvent);
}

ostream& operator<<( ostream& os, InputPort& port )
{
  VectorPEventIt ppEvtIt;

  for ( ppEvtIt=port.m_vecPEvent.begin(); ppEvtIt!=port.m_vecPEvent.end(); ppEvtIt++)
  {
    os << **ppEvtIt;
  }
  os << "\n";
  return os;
}
