// Sim.cpp: implementation of the Sim class.
//
//////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "Sim.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Sim::Sim(char sCfgFileName[], char sOutputFileName[])
{
  m_curTime = 0;
  m_outputTime = 0;
  ReadConfig(sCfgFileName);

  Run();

  if ((sOutputFileName!=NULL) && (sOutputFileName[0]!=0))
  {
    m_ofOutputFile.open(sOutputFileName, ios::app);
    if (!m_ofOutputFile)
      Error("Open %s failed!",sOutputFileName);
    m_ofOutputFile << *this;
  }
  else
    cout << *this;

}

Sim::~Sim()
{
  VectorPSwitchIt ppSwitchIt;
  
  for ( ppSwitchIt=m_vecPSwitch.begin(); 
        ppSwitchIt!=m_vecPSwitch.end(); 
        ppSwitchIt++ )
  {
    delete *ppSwitchIt;
  }
}

// run the simulation on all switches
void Sim::Run()
{
  PEvent pEvent;
  VectorPSwitchIt begin, end;
  time_t preTime, interval, beginTime;
  int i=0;

  beginTime = preTime = time(NULL);
  interval = 30; // seconds
  begin = m_vecPSwitch.begin(); 
  end = m_vecPSwitch.end(); 

  /*
  for (int i=0; i<100000; i++)
  {
    pEvent = m_inputPort.GetNextEvent();
	  delete pEvent;
  }
  /*/
  //*
  VectorPSwitchIt it;
  while ( m_curTime < m_maxTime )
  {
    if (time(NULL)-preTime >= interval)
    {
      char line[100];
      int elapsedTime;
      preTime = time(NULL);
      elapsedTime = preTime-beginTime;
      sprintf(line, "%d:%02d current: %.1lf sec   max: %.1lf sec\n",
              elapsedTime/60, elapsedTime%60, 
              m_curTime/1e9, m_maxTime/1e9);
      cout << *this;
      cout << line;
      cout << flush;
      
      for (it=begin; it!=end; it++)
      {
        if (((*it)->m_bufSize==1000)&&((*it)->m_switchType!=WFQ) &&
            ((*it)->m_switchType!=FIFO_BUF) &&
            ((*it)->m_switchType!=FIFO) &&
            ((*it)->m_method!=PRE_SET))
          (*it)->DebugPrint();
      }
      Check();
    }

    i++;
    pEvent = m_inputPort.GetNextEvent();
    if (pEvent == NULL) // all flows finish
      break;
    assert(m_curTime <= pEvent->pPacket->m_time);
    m_curTime = pEvent->pPacket->m_time;

    //if (i%100==0)
	  //  (*begin)->DebugPrint();
    // output packets
    while ( m_outputTime + m_outputInterval < m_curTime )
    {
      m_outputTime += m_outputInterval;
      for (it=begin; it!=end; it++)
        (*it)->PacketOut(m_outputTime);
    }
    for (it=begin; it!=end; it++)
      (*it)->HandleEvent(pEvent);
	
    delete pEvent;
  }
  //*/
}

void Sim::ReadConfig(char sFileName[])
{
  ifstream cfgFile;
  char line[255];
  double checkInterval;
  double bufSize;
  double headroom;
  double maxTimeSec, warmUp, outputRate;
  int i, numQueue, method;
  char name[20], methodName[20];
  istrstream *pistr;
  BOOL bSwitchEnd;

  cfgFile.open(sFileName, ios::in);
  if (!cfgFile)
    Error("Config file %s not exist!", sFileName);
  ReadLine(cfgFile,line,254);
  // general section
  if (!strstr(line,"[General]"))
    Error("[General] section expected.");

  ReadLine(cfgFile,line,254);
  i = sscanf(line, "CellSize= %d bytes", &CELL_SIZE);
  if (i!=1)
    Error("\"CellSize= # bytes\" expected in config file!");

  ReadLine(cfgFile,line,254);
  i = sscanf(line, "MaxTime= %lf s", &maxTimeSec);
  if (i!=1)
    Error("\"MaxTime= #.# s\" expected in config file!");
  m_maxTime = INT64(maxTimeSec * 1e9);

  ReadLine(cfgFile,line,254);
  i = sscanf(line, "Warmup= %lf s", &warmUp);
  if (i!=1)
    Error("\"Warmup= #.# s\" expected in config file!");

  ReadLine(cfgFile,line,254);
  i = sscanf(line, "OutputRate= %lf Mb/s", &outputRate);
  if (i!=1)
    Error("\"OutputRate= #.# Mb/s\" expected in config file!");
  m_outputInterval = DWORD(TimeScale*8*CELL_SIZE / outputRate / 1e6);

  ReadLine(cfgFile,line,254);
  i = sscanf(line, "CheckInterval= %lf s", &checkInterval);
  if (i!=1)
    Error("\"CheckInterval= #.# s\" expected in config file!");

  ReadLine(cfgFile,line,254);
  i = sscanf(line, "BufSize= %lf Mbytes", &bufSize);
  if (i!=1)
    Error("\"BufSize= #.# Mbytes\" expected in config file!");

  ReadLine(cfgFile,line,254);
  i = sscanf(line, "Headroom= %lf OfBufferSize", &headroom);
  if (i!=1)
    Error("\"Headroom= #.# Mbytes\" expected in config file!");

  // switch section
  ReadLine(cfgFile,line,254);
  if (!strstr(line,"[Switch]")) 
    Error("[Switch] section expected.");

  VectorDouble vecBufSize;
  VectorDoubleIt pVecItBufSize;

  ReadLine(cfgFile,line,254);
  pistr = new istrstream(line);
  *pistr >> name;
  if (stricmp(name,"BufferSize"))
    Error("BufferSize expected");
  while (TRUE)
  {
    *pistr >> bufSize;
    if (bufSize < 0.0)
      break;
    vecBufSize.push_back(bufSize);
  }
  delete pistr;

  bSwitchEnd = FALSE;
  while (!bSwitchEnd)
  {
    ReadLine(cfgFile,line,254);
    pistr = new istrstream(line);
    *pistr >> name;
    if (!stricmp(name,"FIFO_BUF")) // FIFO with buffer management
    {
      for (pVecItBufSize=vecBufSize.begin();
           pVecItBufSize!=vecBufSize.end();
           pVecItBufSize++)
      {
        bufSize = *pVecItBufSize;
        m_vecPSwitch.push_back(new Switch(FIFO_BUF, warmUp, bufSize, 
                              headroom*bufSize, m_outputInterval, 
                              UINT_MAX, checkInterval)); // FIFO switch w buf management
      }
    }
    else if (!stricmp(name,"FIFO")) // FIFO w/o buffer management
    {
      for (pVecItBufSize=vecBufSize.begin();
           pVecItBufSize!=vecBufSize.end();
           pVecItBufSize++)
      {
        bufSize = *pVecItBufSize;
        m_vecPSwitch.push_back(new Switch(FIFO, warmUp, bufSize, 
                              headroom*bufSize, m_outputInterval,
                              1, checkInterval)); // FIFO switch
      }
    }
    else if (!stricmp(name,"WFQ"))
    {
      for (pVecItBufSize=vecBufSize.begin();
           pVecItBufSize!=vecBufSize.end();
           pVecItBufSize++)
      {
        bufSize = *pVecItBufSize;
        m_vecPSwitch.push_back(new Switch(WFQ, warmUp, bufSize, 
                       headroom*bufSize, m_outputInterval, 
                       UINT_MAX, checkInterval)); // WFQ switch
      }
    }
    else if (!stricmp(name,"Hybrid"))
    {
      *pistr >> numQueue >> methodName;
      
      method = -1;
      if (!stricmp(methodName,"PRE_SET"))
        method = PRE_SET;
      else if (!stricmp(methodName,"RATE_BALANCE"))
        method = RATE_BALANCE;
      else if (!stricmp(methodName,"BURST_BALANCE"))
        method = BURST_BALANCE;
      else if (!stricmp(methodName,"SIMILAR"))
        method = SIMILAR;
      else if (!stricmp(methodName,"RANDOM"))
        method = RANDOM;
      else if (!stricmp(methodName,"ROUND_ROBIN"))
        method = ROUND_ROBIN;
      else if (!stricmp(methodName,"ADAPTIVE"))
        method = ADAPTIVE;
      else if (!stricmp(methodName,"NEAR_OPTIMAL"))
        method = NEAR_OPTIMAL;
      else if (!stricmp(methodName,"OPTIMAL"))
        method = OPTIMAL;
      else
        Error("Error hybrid method in config!");

      for (pVecItBufSize=vecBufSize.begin();
           pVecItBufSize!=vecBufSize.end();
           pVecItBufSize++)
      {
        bufSize = *pVecItBufSize;
        m_vecPSwitch.push_back(new Switch(Hybrid, warmUp, bufSize, headroom*bufSize, 
             m_outputInterval, numQueue, checkInterval, method, methodName)); // Hybrid switch
      }
    }
    else if (!stricmp(name, "End"))
      bSwitchEnd = TRUE;
    else
      Error("Unknown switch kind!");

    delete pistr;
  }  //  while (!bSwitchEnd)

  // flow section
  ReadLine(cfgFile,line,254);
  if (!strstr(line,"[Flow]")) 
    Error("[Flow] section expected.");

  if (strstr(line, "name")) // use the specified file
  {
    char s1[20],s2[20],s3[20];
    cfgFile.close();
    sscanf(line, "%s %s %s", s1, s2, s3);
    cfgFile.open(s3);
    if (!cfgFile)
      Error("Open %s error!", s3);
  }
  m_inputPort.InitVc(cfgFile);
}

ostream& operator<<( ostream& os, Sim &s )
{
  size_t i;

  os << "Name          Buf   Headrm Thr    AggTh   ConTh  Con    ModTh  Mod     Benifit\n";
  os << "              MB    MBytes Mb/s   Mb/s    Mb/s   Los    Mb/s   Los\n";
  for (i=0; i<s.m_vecPSwitch.size(); i++)
    os << *(s.m_vecPSwitch[i]);
  return os;
}

void Sim::Check()
{
  VectorPSwitchIt it;
  PSwitch pSwitch;
  VectorPSwitchIt begin, end;

  begin = m_vecPSwitch.begin(); 
  end = m_vecPSwitch.end(); 
  for (it=begin; it!=end; it++)
  {
    pSwitch = *it;
    pSwitch->Check();
  }
}
