// Switch.h: interface for the Switch class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(_SWITCH_H_INCLUDED_)
#define _SWITCH_H_INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "global.h"	// Added by ClassView
#include "Packet.h"
#include "Event.h"
#include "VC.h"
#include "VCHash.h"
#include "Event.h"

typedef list<PPacket> QueuePPacket;
typedef QueuePPacket::iterator QueuePPacketIt;
typedef QueuePPacket* PQueuePPacket;

typedef class QueueHybrid {
public:
  QueuePPacket m_quePPacket;
  ListPVC m_listPVc;
  ListPVC m_listPVcPast;
  int m_id;
  DWORD m_threshold; 
  double m_tokenRate;  // the sum of all tokenRate of vcs in this hybrid queue
  double m_burstSize;
  INT64 m_tick;       // m_vClock increase m_tick every time.  m_tick = 1/m_tokenRate
  INT64 m_vClock;     // the queue with the lowest vClock will be outputed
  DWORD m_total;
  DWORD m_outNum;     // number of cells outputed until now after warmup
  DWORD m_loss;
  DWORD m_allTotal;
  DWORD m_allOutNum;     // number of cells outputed until now
  DWORD m_allLoss;
public:
	void Check(size_t &vcnum);
  QueueHybrid();
  virtual ~QueueHybrid();
  friend ostream& operator<<( ostream& os, QueueHybrid &q );
}* PQueueHybrid;

typedef vector<QueueHybrid> VectorQueueHybrid;
typedef VectorQueueHybrid::iterator VectorQueueHybridIt;

typedef vector<PQueueHybrid> VectorPQueueHybrid;
typedef VectorPQueueHybrid::iterator VectorPQueueHybridIt;

enum
{
  NONE,
  PRE_SET,    // select a queue with the same id as indicated
  RATE_BALANCE,   // select a queue with smallest token rate
  BURST_BALANCE,   // select a queue with smallest burst size
  SIMILAR,    // select a queue with similar traffic profile
  RANDOM,     // select a random queue
  ROUND_ROBIN,// select in a sequence of round robin
  NEAR_OPTIMAL,     // use optimal algorithm
  ADAPTIVE,   // shaffle queue dynamicly
  OPTIMAL     // use optimal algorithm
};

typedef struct Flow
{
  int m_id;
  double m_burst;
  double m_rate;
}* PFlow;
typedef list<Flow> ListFlow;

typedef struct Queue
{
  ListFlow m_listFlow;
  double m_burst;
  double m_rate;
}* PQueue;
typedef vector<Queue> VectorQueue;

typedef class Switch  
{
public:
  //DWORD m_maxLen[1000]; // for debug
  //DWORD m_max;

  int m_switchType;
  char m_name[20];        // name of this switch
  DWORD m_bufSize;        // total buffer size in cells of this switch
  DWORD m_totalHeadroom;  // total number of reserved buffer
  DWORD m_headroom;       // current number of free headroom
  DWORD m_totalHoles;     // total number of buffer for sharing
  DWORD m_holes;          // current number of buffer for sharing
  DWORD m_maxQueue;       // the max number of hybrid queues
  DWORD m_outputInterval; // the time needed to output a packet in cells/ns
  DWORD m_outNum;     // number of cells outputed until now after warmup
  DWORD m_allOutNum;     // number of cells outputed until now
  DWORD m_total;        // number of cells inputed after warmup
  DWORD m_allTotal;        // number of cells inputed
  DWORD m_loss;         // number of cells lossed
  DWORD m_allLoss;         // number of cells lossed
  /*
  // information of finished vc
  DWORD m_pastAllTotal;
  DWORD m_pastAllOutNum;
  DWORD m_pastAllLoss;
  // information of finished vc after warmup
  DWORD m_pastTotal;
  DWORD m_pastOutNum;
  DWORD m_pastLoss;
  */
  VCHash m_vcHash;      // hash table for vc
  int m_method;         // the method to apply a vc to a hybrid queue
  
  // vector of the hybrid queues, orgnized in heap in order to compute the 
  // smallest virtual clock
  VectorPQueueHybrid m_vecPQueHybrid; 
  INT64 m_interval;   // adjust vClock every m_interval time
  INT64 m_checkTime;  // the last time that the vClock was checked
  INT64 m_curTime;
  INT64 m_warmUp;
  RandGen m_rand;     // for SelectRandom()
  size_t m_curRRIndex;  // for SelectRoundRobin()
  INT64 m_prevTime;
public:
	void Check();
	void CheckVClock(INT64 curTime);
	void HandleEvent(PEvent pEvent); // a packet enters the switch
	void PacketOut(INT64 curTime); // output a packet
  double GetOptBenifit(size_t k=0, double burst=0.0, double rate=0.0);
	void DebugPrint();

	Switch(int switchType, double warmUp, 
         double bufSize, double totalHeadroom, // MBytes
         DWORD outputInterval, int maxQueue, 
         double checkInterval, int method=NONE, char methodName[]=NULL);
	virtual ~Switch();
protected:
	static BOOL LaterVirtualClock(PQueueHybrid &q1, PQueueHybrid &q2);
	PVC OutputPacket(PQueueHybrid pQue);
	void AdjustQueueThreshold();
  // determine how to accept a new vc
  void AcceptVc(PEvent pEvent);
	void PrintThreshold();
	void TraceAll();
	PQueueHybrid SelectQueue(PEvent pEvent);
	PQueueHybrid SelectMinRate(PEvent pEvent);
	PQueueHybrid SelectMinBurst(PEvent pEvent);
	PQueueHybrid SelectPreSet(PEvent pEvent);
  PQueueHybrid SelectSimilar(PEvent pEvent);
  PQueueHybrid SelectRandom(PEvent pEvent);
  PQueueHybrid SelectRoundRobin(PEvent pEvent);
  PQueueHybrid SelectAdaptive(PEvent pEvent);
  PQueueHybrid SelectNearOptimal(PEvent pEvent);
  PQueueHybrid SelectOptimal(PEvent pEvent);
  PQueueHybrid SelectSimilarFromRange(PEvent pEvent,int begin, int end,
                                      double &minDiff);
  friend ostream& operator<<( ostream& os, Switch &s );
}* PSwitch;

// switch type
enum
{
  FIFO = 1,   // FIFO w/o buffer management
  FIFO_BUF, // FIFO with buffer sharing
  WFQ,
  Hybrid
};

typedef vector<Switch> VectorSwitch;
typedef VectorSwitch::iterator VectorSwitchIt;
typedef vector<PSwitch> VectorPSwitch;
typedef VectorPSwitch::iterator VectorPSwitchIt;

ostream& operator<<( ostream& os, Switch &s );

#endif // !defined(_SWITCH_H_INCLUDED_)
