// VC.h: interface for the VC class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(_VC_H_INCLUDED_)
#define _VC_H_INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

typedef class VC* PVC;
typedef class QueueHybrid* PQueueHybrid;
typedef class Event* PEvent;

#include "Packet.h"
#include "RandGen.h"
#include "global.h"	// Added by ClassView

// information of a virtual channel
typedef class VC  
{
public:
	PEvent GetNextEvent();
public:
	VC(int id, int qid,
       double beginTimeSec, double lengthSec, // seconds
       double peakRate, double avgRate, double tokenRate, // Mbits/s
       double tokenBucket, // KBytes
       int type,
       double onTime, // seconds
       int seedOn, int seedOff);
	virtual ~VC();
public:
	void SetSeed(int seedOn, int seedOff);
  int m_id;
  int m_qid;

  // used by Switch
  DWORD m_outNum;     // number of cells outputed until now after warmup
  DWORD m_allOutNum;     // number of cells outputed until now
  DWORD m_num;     // current number of cells in the output queue
  DWORD m_loss;    // number of cells that are lost
  DWORD m_allLoss;    // number of cells that are lost
  PQueueHybrid m_pQueHybrid; // the pointer to the HybridQueue which contains this vc
  DWORD m_flag;    // indicate whether the vc stops
  DWORD m_total;   // total cells of this vc until now after warmup
  DWORD m_allTotal;   // total cells of this vc until now
  DWORD m_tokenInterval;     // ns / cell
  double m_burstSize;     // cells
  BOOL m_conform;       // whether this stream is conformant
  int m_type;
  // peakRate and avgRate may exceed tokenRate and tokenBucket for aggressive streams
  DWORD m_peakInterval;      // ns / cell
  double m_avgRate;     // cells / ns = m_peakRate * m_avgOn / (m_avgOn+m_avgOff)

protected:
  static int m_seq;         // sequence number of packets
  // used by InputPort to generate new cells
  INT64 m_beginTime;    // begin time of this vc
  INT64 m_endTime;      // end time of this vc

  // double m_peakRate;     // cells / ns = 1/m_peakInterval
  DWORD m_tokenBucket;       // cells
  // double m_tokenRate;    // cells / ns = 1/m_tokenInterval
  INT64 m_avgOn;         // ns
  INT64 m_avgOff;        // ns
  RandGen m_randOn;
  RandGen m_randOff;

  INT64 m_prevCell;        // arrival time of the previous cell
  INT64 m_prevToken;          // arrival time of the previous token
  DWORD m_waiting;
  BOOL m_on;
  INT64 m_onBeginTime;
  INT64 m_onEndTime;     // the end time of the current on period
  DWORD m_curBucket;      // number of cells waiting in the bucket
  int  m_prevEvt;       // last event generated
  INT64 m_prevTime;     // time of last event
protected:
	INT64 NextPossionEvent(RandGen &randGen, DWORD avgInterval);
  INT64 OnPeriod();
  INT64 OffPeriod();
}* PVC;

typedef vector<VC> VectorVC;
typedef list<VC> ListVC;
typedef list<PVC> ListPVC;
typedef vector<ListVC> VectorListVC;
typedef ListVC::iterator ListVCIt;
typedef ListPVC::iterator ListPVCIt;
typedef VectorListVC::iterator VectorListVCIt;

#endif // !defined(_VC_H_INCLUDED_)
