// VcGen.cpp : Defines the entry point for the console application.
//

#include "StdAfx.h"
#include "RandGen.h"

typedef struct Vc
{
  double m_beginTime;
  double m_endTime;
  int m_id;
  double m_peakRate;
  double m_avgRate;
  double m_tknRate;
  double m_tknBckt;
  int m_type;
  double m_avgOn;
  int m_onSeed;
  int m_offSeed;
  double m_lenOn;
  double m_lenOff;
  int m_repeat;
  RandGen m_randOn;
  RandGen m_randOff;
}* PVc;

double NextPossionEvent(RandGen &randGen, double interval)
{
  double rand;
  double r;

  rand = double(randGen.random()) / RAND_M;
  r = -log(rand)*interval;
  assert(r>=0);
  return r;
}

void Error(char *fmt, ...)
{
  va_list args;

  cerr << "Error: ";

  va_start(args, fmt);
  vfprintf(stderr, fmt, args);
  fprintf(stderr, "\n");
  va_end(args);
    
  exit(1);
}

BOOL IsBlank(char line[])
{
  char *pch;
  pch = line;
  while ((*pch==' ')||(*pch=='\t'))
    pch++;
  if ((*pch==0)||(*pch=='#'))
    return TRUE;
  else
    return FALSE;
}

// read a line, ignore comment lines and empty lines
void ReadLine(ifstream &f, char line[], int len)
{
  do
  {
    f.getline(line,len);
  }while (IsBlank(line) && !f.eof());
}

void WriteVc(ofstream &of, PVc pVc)
{
  char line[255];

//Qid PeakRate AvgRate TknBckt TknRate Type AvgOn   On  Off BeginTime Length Repeat
//    Mbits/s  Mbits/s KBytes  Mbits/s          sec   Seed Seed second   second
  sprintf(line, "%-4d %lf %lf %lf %lf %d %lf %5d  %5d %f %f %3d\n",
                pVc->m_id, pVc->m_peakRate, pVc->m_avgRate, pVc->m_tknBckt,
                pVc->m_tknRate, pVc->m_type, pVc->m_avgOn, pVc->m_onSeed, pVc->m_offSeed,
                pVc->m_beginTime, pVc->m_endTime-pVc->m_beginTime, 1);
  of << line;
}

int main(int argc, char* argv[])
{
  PVc pVc;
  ifstream ifCfgIn;
  ofstream ofCfgOut;
  char line[255];
  istrstream *pis;
  int i, j;
  double maxTime;
  int id;

  if (argc!=3)
    Error("Usage: VcGen <inputfile> <outputfile>");
  ifCfgIn.open(argv[1]);
  if (!ifCfgIn)
    Error("Open %s fails!",argv[1]);

  ofCfgOut.open(argv[2]);
  if (!ofCfgOut)
    Error("Open %s fails!",argv[2]);
  ofCfgOut << "#Qid PeakRate AvgRate TknBckt TknRate Type AvgOn   On  Off BeginTime Length Repeat\n";
  ofCfgOut << "#    Mbits/s  Mbits/s KBytes  Mbits/s       sec   Seed Seed second   second\n";

  ReadLine(ifCfgIn,line,254);
  // general section
  if (!strstr(line,"[General]"))
    Error("[General] section expected.");

  ReadLine(ifCfgIn,line,254);
  i = sscanf(line, "MaxTime= %lf s", &maxTime);
  if (i!=1)
    Error("\"MaxTime= #.# s\" expected in config file!");

  ReadLine(ifCfgIn,line,254);
  // flow section
  if (!strstr(line,"[Flow]"))
    Error("[General] section expected.");

  id = 0;
  while (!ifCfgIn.eof())
  {
    ReadLine(ifCfgIn, line, 254);
    if (IsBlank(line) && ifCfgIn.eof())
      break;

    pVc = new Vc;
    pVc->m_beginTime = pVc->m_endTime = 0.0;
    pis = new istrstream(line);
//Qid PeakRate AvgRate TknBckt TknRate Conform AvgOn   On  Off BeginTime Length Repeat
//    Mbits/s  Mbits/s KBytes  Mbits/s          sec   Seed Seed second   second
    *pis >> pVc->m_id >> pVc->m_peakRate >> pVc->m_avgRate 
         >> pVc->m_tknBckt >> pVc->m_tknRate >> pVc->m_type 
         >> pVc->m_avgOn >> pVc->m_onSeed >> pVc->m_offSeed
         >> pVc->m_lenOn >> pVc->m_lenOff >> pVc->m_repeat;
    delete pis;

    if (pVc->m_repeat<1)
      Error("Config file repeat error!");

    for (i=0; i<pVc->m_repeat; i++)
    {
      pVc->m_beginTime = pVc->m_endTime = 0.0;
      pVc->m_randOn.srandom(((i+1)*177+9119)*((i+1)*177+9119));
      pVc->m_randOff.srandom(((i+1)*97+9119)*((i+1)*97+9119));
      for (j=0; j<100; j++)
      {
        pVc->m_randOn.random();
        pVc->m_randOff.random();
      }
      while (TRUE)
      {
        pVc->m_id = id++;
        pVc->m_onSeed = abs((INT64((id+51)*139+8537))*((id+51)*139+8537)/100%10000)+id;
        pVc->m_offSeed = abs((INT64((id+47)*203+7901))*((id+47)*203+7901)/100%10000)+id;
        pVc->m_beginTime = pVc->m_endTime + NextPossionEvent(pVc->m_randOff, pVc->m_lenOff);
        assert(pVc->m_beginTime>=pVc->m_endTime);
        pVc->m_endTime = pVc->m_beginTime + NextPossionEvent(pVc->m_randOn, pVc->m_lenOn);
        assert(pVc->m_endTime>=pVc->m_beginTime);
        if (pVc->m_beginTime > maxTime)
          break;
        WriteVc(ofCfgOut, pVc);
      }
    }
    delete pVc;
  }
  cout << '\a';
	return 0;
}
